/* Project 1 --- EE 368 --- Fall 2000 */

/* Team Project: ee368t16
 * Member : ee368d11
 *          ee368c13
 *          ee368d14
 */

#include "DPM.h"


void Insert_In_Wait_Queue (
  Wait_Queue *Customer_Wait_Queue,
  Customer_Pointer Ptr_Customer)
{
  /*  The wait queue is maintained as a singly linked 
   *  list with pointers to both the head and tail of the list.
   *  The list is empty if both the head and tail pointers are NULL.
   *  If the list is not empty then the next field of the tail
   *  node is NULL.
   * 
   *  The pointer to the customer is inserted into a new node
   *  which is placed at the tail end of the wait list and the
   *  wait queue length is increased by one.
   *
   *  The number of items in the queue is increased by the the
   *  number of items purchased by the customer.
   */

  Wait_Node_Pointer New_Wait_Node_Pointer = NULL;

  New_Wait_Node_Pointer = (Wait_Node_Pointer) malloc (sizeof(Wait_Node));

  if (New_Wait_Node_Pointer == NULL)
  {
    exit ERROR_DETECTED;
  } 

  New_Wait_Node_Pointer->Ptr_Customer = Ptr_Customer;
  New_Wait_Node_Pointer->Ptr_Next = NULL;

  /* Check for empty Customer_Wait_Queue */

  if (    ((*Customer_Wait_Queue).Ptr_Head == NULL)
       && ((*Customer_Wait_Queue).Ptr_Tail == NULL))

  {
    (*Customer_Wait_Queue).Ptr_Head = New_Wait_Node_Pointer;
    (*Customer_Wait_Queue).Ptr_Tail = New_Wait_Node_Pointer;
  }
  else
  {
    /*
     * Connect the new node to the end of
     * Customer_Wait_Queue
     */

    (*Customer_Wait_Queue).Ptr_Tail->Ptr_Next = New_Wait_Node_Pointer;
    (*Customer_Wait_Queue).Ptr_Tail = New_Wait_Node_Pointer;
  }

  (*Customer_Wait_Queue).Wait_Queue_Length++;
  (*Customer_Wait_Queue).Wait_Items_In_Queue =
     (*Customer_Wait_Queue).Wait_Items_In_Queue +
     Ptr_Customer->Parameters.Number_Of_Items_Purchased;

  return;

} /* Insert_In_Wait_Queue */



Customer_Pointer Remove_From_Wait_Queue (
  Wait_Queue *Customer_Wait_Queue)
{
  /*  The wait queue is maintained as a singly linked 
   *  list with pointers to both the head and tail of the list.
   *  The list is empty if both the head and tail pointers are NULL.
   *  If the list is not empty then then the next field of the tail
   *  node is NULL.
   * 
   *  The node at the head of the list is removed and
   *  the pointer to the next waiting customer stored
   *  in that node is returned to the calling routine
   *  and then the node is freed.
   *
   *  The wait queue legth is decreased by one.
   *  The number of items in the wait queue is decreased by the
   *  number of items purchased by the customer just removed 
   *  from the list.
   */

  Customer_Pointer Ptr_Customer = NULL;
  Wait_Node_Pointer Node_To_Be_Removed = NULL;

  /* Check for empty list */

  if (    ((*Customer_Wait_Queue).Ptr_Head == NULL)
       && ((*Customer_Wait_Queue).Ptr_Tail == NULL))
  {
    exit ERROR_DETECTED;
  }
  else
  {
    Ptr_Customer = (*Customer_Wait_Queue).Ptr_Head->Ptr_Customer;

    /* Check for single Node in the wait queue */

    if ((*Customer_Wait_Queue).Ptr_Head->Ptr_Next == NULL)
    {
      (*Customer_Wait_Queue).Ptr_Head = NULL;
      (*Customer_Wait_Queue).Ptr_Tail = NULL; 
    }
    else
    {
      /* Multi node list condition */

      Node_To_Be_Removed = (*Customer_Wait_Queue).Ptr_Head;      
      (*Customer_Wait_Queue).Ptr_Head = 
         (*Customer_Wait_Queue).Ptr_Head->Ptr_Next; 

      free (Node_To_Be_Removed);
      Node_To_Be_Removed = NULL;
    }
  }

  (*Customer_Wait_Queue).Wait_Queue_Length--;
  (*Customer_Wait_Queue).Wait_Items_In_Queue =
     (*Customer_Wait_Queue).Wait_Items_In_Queue -
     Ptr_Customer->Parameters.Number_Of_Items_Purchased;
  
  return Ptr_Customer;

} /* Remove_From_Wait_Queue */


void Insert_In_Next_Event_List (
  Event_Node_Pointer *Ptr_Head,
  int Customer_Event_Time,
  Events Customer_Status,
  Customer_Pointer Ptr_Customer)
{
  /*  The next event list is maintained as a singly linked 
   *  list with a pointer to the head of the list.
   *  The list is empty of the pointer to the head is NULL.
   *  If the list is not empty then the next field of last
   *  item in the list is NULL.
   * 
   *  The customer event time, customer status, and
   *  a pointer to the customer, are stored in a new node
   *  which is inserted in the next event list in it proper
   *  place. The events are sorted from head to tail with
   *  the lowest customer event times toward the head.
   *  In case of event time ties the new node is inserted after
   *  all nodes with the same event time (toward the tail).
   */

  Event_Node_Pointer New_Event_Node_Pointer = NULL;

  Event_Node_Pointer Temp_1 = NULL;
  Event_Node_Pointer Temp_2 = NULL;

  New_Event_Node_Pointer = (Event_Node_Pointer) malloc
                              (sizeof(Event_Node));

  if (New_Event_Node_Pointer == NULL)
  {
    exit (ERROR_DETECTED);
  }

  /*
   * Setting up the new node with the customer event time,
   * the customer status and a pointer to the customer
   */

  New_Event_Node_Pointer->Ptr_Customer = Ptr_Customer;
  New_Event_Node_Pointer->Customer_Event_Time =
                              Customer_Event_Time;
  New_Event_Node_Pointer->Customer_Status = Customer_Status;
  New_Event_Node_Pointer->Ptr_Next = NULL;

  if ((*Ptr_Head) == NULL)
  { 
    /* The list is empty */
    
    (*Ptr_Head) = New_Event_Node_Pointer;
  }
  
  else if ((*Ptr_Head)->Ptr_Next == NULL)
  {
    /* Single Node List */    

    if (New_Event_Node_Pointer->Customer_Event_Time <
            (*Ptr_Head)->Customer_Event_Time)
    {    
      /* Insert at the begining of the List */
      
      New_Event_Node_Pointer->Ptr_Next = (*Ptr_Head);
      (*Ptr_Head) = New_Event_Node_Pointer;
    }
    else if (New_Event_Node_Pointer->Customer_Event_Time >=
          (*Ptr_Head)->Customer_Event_Time)
    {
      /* Insert at the end of the List */  

      (*Ptr_Head)->Ptr_Next = New_Event_Node_Pointer;
    }      
  }
  else if ((*Ptr_Head)->Ptr_Next != NULL)
  {
    /* Multiple Node */
    
    if (New_Event_Node_Pointer->Customer_Event_Time <
         (*Ptr_Head)->Customer_Event_Time)
    {
      New_Event_Node_Pointer->Ptr_Next = (*Ptr_Head);
      (*Ptr_Head) = New_Event_Node_Pointer;
    }  
    
    else if (New_Event_Node_Pointer->Customer_Event_Time >=
              (*Ptr_Head)->Customer_Event_Time)
    {
      Temp_1 = (*Ptr_Head)->Ptr_Next;     
      Temp_2 = (*Ptr_Head);     
      
      while (    (New_Event_Node_Pointer->Customer_Event_Time >=
                      Temp_1->Customer_Event_Time)
              && (Temp_1->Ptr_Next != NULL))
      {
        Temp_1 = Temp_1->Ptr_Next;
        Temp_2 = Temp_2->Ptr_Next;
      }   
      if (New_Event_Node_Pointer->Customer_Event_Time >=
        Temp_1->Customer_Event_Time)
      {
        Temp_1->Ptr_Next = New_Event_Node_Pointer;
      }
      else
      {
        Temp_2->Ptr_Next = New_Event_Node_Pointer;
        New_Event_Node_Pointer->Ptr_Next = Temp_1;
      }
    }
  }
  Temp_1 = NULL;
  Temp_2 = NULL;

  return;

} /* Insert_In_Next_Event_List */

Customer_Pointer Remove_From_Next_Event_List (
  Event_Node_Pointer *Ptr_Head,
  int *Customer_Event_Time,
  Events *Customer_Status)
{
  /*  The next event list is maintained as a singly linked 
   *  list with a pointer to the head of the list.
   *  an empty list has the head pointer equal to NULL.
   *  If the list is not empty then the next field of the
   *  last node points to NULL.
   * 
   *  The node at the head of the list is removed and the
   *  customer event time, customer status, and the pointer
   *  to the customer stored in the node are passed back to
   *  the calling routine.  The removed node is then freed.
   */

  Customer_Pointer Ptr_Customer = NULL;
  Event_Node_Pointer Node_To_Be_Removed = NULL;

  /* Check for empty list */
    
  if ((*Ptr_Head) == NULL)
  {
    exit ERROR_DETECTED;
  }
 
  /* Check for single node list */

  if ((*Ptr_Head)->Ptr_Next == NULL)
  {   
    Ptr_Customer = (*Ptr_Head)->Ptr_Customer;    
    *Customer_Event_Time = (*Ptr_Head)->Customer_Event_Time;
    *Customer_Status = (*Ptr_Head)->Customer_Status;

    free (*Ptr_Head);
    (*Ptr_Head) = NULL;     
  }
  else
  {
    /* Multiple Node list */
  
    Node_To_Be_Removed = (*Ptr_Head);
    Ptr_Customer = (*Ptr_Head)->Ptr_Customer;
    *Customer_Event_Time = Node_To_Be_Removed->Customer_Event_Time;
    *Customer_Status = Node_To_Be_Removed->Customer_Status;
    (*Ptr_Head) = (*Ptr_Head)->Ptr_Next;

    /* Freeing the node on the Head of the event queue */

    free (Node_To_Be_Removed);
    Node_To_Be_Removed = NULL;
   }     

  return Ptr_Customer;

} /* Remove_From_Next_Event_List */



void Handle_New_Customer (
  Customer_Pointer Ptr_Customer,
  Event_Node_Pointer *Ptr_Next_Event_List,
  int Event_Time)
{
  /*
   *  A pointer to the new customer is inserted in the next
   *  event list with the customer event time set to the current
   *  event time + the customer interarrival time.
   *  The customer status is set to Arrived
   */

  int Customer_Event_Time = 0;
  Events Customer_Status = Invalid;

  Customer_Event_Time = Event_Time +
     Ptr_Customer->Parameters.Interarrival_Time;
  Customer_Status = Arrived;

  Insert_In_Next_Event_List (
    &(*Ptr_Next_Event_List),
    Customer_Event_Time,
    Customer_Status,
    Ptr_Customer);

  return;

} /* Handle_New_Customer */


void Handle_Arrived (
  Customer_Pointer Ptr_Customer,
  int Event_Time,
  Event_Node_Pointer *Ptr_Next_Event_List)
{
  int Local_Customer_Event_Time = 0;
  Events Local_Customer_Status = Invalid;

  /*
   *  The customer arrived time is set to the current event time.
   */

  Ptr_Customer->Time_Arrived = Event_Time;

  /*
   *  If the number of items purchased by the customer is greater
   *  than zero.
   */

  if (Ptr_Customer->Parameters.Number_Of_Items_Purchased > 0)
  {

    /* 
     * The customer event time set to the current
     * event time + the customer shopping time.
     */

    Local_Customer_Event_Time = (Event_Time 
        + Ptr_Customer->Parameters.Shopping_Time);

    /*  The customer status is set to Checking_Out  */

    Local_Customer_Status = Checking_Out;
    
    /*  A pointer to the customer is inserted in the next event list  */

    Insert_In_Next_Event_List (
        &(*Ptr_Next_Event_List),
        Local_Customer_Event_Time,
        Local_Customer_Status,
        Ptr_Customer);     
  }
  else
  {
    Local_Customer_Event_Time = (Event_Time 
        + Ptr_Customer->Parameters.Shopping_Time);
    Local_Customer_Status = Departing;

    Insert_In_Next_Event_List (
        Ptr_Next_Event_List,
        Local_Customer_Event_Time,
        Local_Customer_Status,
        Ptr_Customer);  
  }
  
  return;

} /* Handle_Arrived */



void Handle_Checking_Out (
  Customer_Pointer Ptr_Customer,
  int Event_Time,
  Event_Node_Pointer *Ptr_Next_Event_List,
  Wait_Queue Wait_Queue [SERVER_LIMIT + 1],
  Queue_Method Queue_Method,
  Server_State Server_Status [SERVER_LIMIT + 1],
  int Number_Of_Servers)
{
  int Server_Id = 0;
  int Local_Customer_Event_Time = 0;
  int Shortest_Server = 0;
  int Shortest_Item = 0;
  Events Local_Customer_Status = Invalid;
  Boolean Server_Found = False;

  /*
   *  The customer check out time is set to the current event time
   */
 
  Ptr_Customer->Time_Check_Out = Event_Time;

  /*
   *  If there is a server(s) available
   */

  do
  {
    /*
     *  The available server with the lowest server number is
     *  assigned to the customer. The server status is marked busy. 
     */
  
    if (Server_Status[Server_Id + 1] == Available)
    {
      Server_Found = True;
      Ptr_Customer->Server = Server_Id + 1;
      Server_Status[Server_Id + 1] = Busy;
    }

    Server_Id++;
  } while (    (Server_Found == False)
            && (Server_Id < Number_Of_Servers));

  if (Server_Found == True)
  {
    /*  
     * A pointer to the customer is inserted in the next event
     * list with the customer event time set to the current
     * event time + the customer service time.
     *
     * The customer status is set to Departing
     */

    Local_Customer_Event_Time = (Event_Time +
        Ptr_Customer->Parameters.Service_Time);
    Local_Customer_Status = Departing;
    Insert_In_Next_Event_List (Ptr_Next_Event_List,
        Local_Customer_Event_Time,
        Local_Customer_Status,
        Ptr_Customer);
  }

  /*  else if the queue method is a single queue */

  else if (Queue_Method == Single_Queue)
  {
    /*
     *  The queue length field of the customer is set to the
     *  wait queue length + 1.
     */

    Ptr_Customer->Queue_Length = Wait_Queue[0].Wait_Queue_Length + 1;

    /*
     *  The items in queue field of the customer is set to the
     *  number of items in the wait queue plus the number of
     *  items purchased by the customer
     */

    Ptr_Customer->Items_In_Queue = 
          (Wait_Queue[0].Wait_Items_In_Queue +
          Ptr_Customer->Parameters.Number_Of_Items_Purchased);

    
    /*  A pointer to the customer is placed in Wait_Queue[0]. */

    Insert_In_Wait_Queue (&Wait_Queue[0], Ptr_Customer);

  }

  /* else if the queue method is shortest server queue */

  else if (Queue_Method == Shortest_Queue)
  {
    /*
     *  The server with the shortest number of customer is identified.
     *  In case of ties the lowest numbered server is selected.
     */

    for (Server_Id = 0; Server_Id < Number_Of_Servers; Server_Id++)
    {
       if (   (Wait_Queue[Server_Id].Wait_Queue_Length)
            < (Wait_Queue[Shortest_Server].Wait_Queue_Length))
       {
         Shortest_Server = Server_Id;
       }
    }  

    /*
     *  The queue length field of the customer is set to the
     *  wait queue length + 1.
     */

    Ptr_Customer->Queue_Length = 
        Wait_Queue[Shortest_Server].Wait_Queue_Length + 1;

    /*    
     *  The items in queue field of the customer is set to the
     *  number of items in the wait queue plus the number of
     *  items purchased by the customer
     */

    Ptr_Customer->Items_In_Queue = 
          (Wait_Queue[Shortest_Server].Wait_Items_In_Queue +
          Ptr_Customer->Parameters.Number_Of_Items_Purchased); 

    /*  
     *  A pointer to the customer is placed in the wait queue
     *  of the selected server.     
     */

    Insert_In_Wait_Queue (&Wait_Queue[Shortest_Server],
        Ptr_Customer);
  }

  /*  else if the queue method is fewest items queue */

  else if (Queue_Method == Fewest_Item_Queue)
  {
    /*
     *  The server with the fewest number of items is idenified.
     *  In case of ties the lowest numbered server is selected.
     */

    for (Server_Id = 0; Server_Id < Number_Of_Servers; Server_Id++)
    {
      if (   (Wait_Queue[Server_Id].Wait_Items_In_Queue)
           < (Wait_Queue[Shortest_Item].Wait_Items_In_Queue))
      {
          Shortest_Item = Server_Id;
      }
    }
        
    /*
     *  The queue length field of the customer is set to the
     *  wait queue length + 1.
     */

    Ptr_Customer->Queue_Length =
        Wait_Queue[Shortest_Item].Wait_Queue_Length + 1;
    
    /*
     *  The items in queue field of the customer is set to the
     *  number of items in the wait queue plus the number of
     *  items purchased by the customer
     */
          
    Ptr_Customer->Items_In_Queue =
          (Wait_Queue[Shortest_Item].Wait_Items_In_Queue +
          Ptr_Customer->Parameters.Number_Of_Items_Purchased);  
     
    /*  
     *  A pointer to the customer is placed in the wait queue
     *  of the selected server.
     */
       
    Insert_In_Wait_Queue (&Wait_Queue[Shortest_Item], Ptr_Customer);
  }

  return;
} /* Handle_Checking_Out */


void Handle_Departing (
  Customer_Pointer Ptr_Customer,
  int Event_Time,
  Event_Node_Pointer *Ptr_Next_Event_List,
  Wait_Queue Wait_Queue [SERVER_LIMIT + 1],
  Queue_Method Queue_Method,
  Server_State Server_Status [SERVER_LIMIT + 1],
  int Number_Of_Servers)
{
  /*
   * The customer's time departed is set to the current event time
   *
   *  If the server number is not zero, (nothing purchased)
   *
   *    If the queue method is single queue
   *  
   *      If Wait_Queue[0] is not empty
   *        The next customer to be servered is removed from the
   *        Wait_Queue[0].
   *  
   *        The current server is assigned to the new customer. 
   *  
   *        A pointer to the new customer is inserted in the next event
   *        list with the customer event time set to the current
   *        event time + the customer service time.
   *  
   *        The new customer's status is set to Departing
   *  
   *      else
   *   
   *        The server is marked available.
   *  
   *    else if the queue method is either
   *         shortest queue or fewest items
   *  
   *      If the server wait queue is not empty
   *        The next customer to be servered is removed from the
   *        server's wait queue.
   *  
   *        The current server is assigned to the new customer. 
   *  
   *        A pointer to the new customer is inserted in the next event
   *        list with the customer event time set to the current
   *        event time + the customer service time.
   *  
   *        The new customer's status is set to Departing
   *  
   *      else
   *   
   *        The server is marked available.
   */

  int Customer_Event_Time = 0;
  Customer_Pointer Customer_To_Be_Served = NULL;
  Events Customer_To_Be_Served_Status = Invalid;
  int Wait_Queue_Array = 0;

  Ptr_Customer->Time_Departed = Event_Time;

  if (Ptr_Customer->Server != 0)
  {
    if (Queue_Method == Single_Queue)
    {
      if (Wait_Queue[0].Ptr_Head != NULL)
      {
        Customer_To_Be_Served = Remove_From_Wait_Queue (
                                  &(Wait_Queue[0]));

        Customer_To_Be_Served->Server = Ptr_Customer->Server;
        Customer_Event_Time = 
          Event_Time + Customer_To_Be_Served->Parameters.Service_Time;
        Customer_To_Be_Served_Status = Departing;
        
        Insert_In_Next_Event_List (
          &(*Ptr_Next_Event_List),
          Customer_Event_Time,
          Customer_To_Be_Served_Status,
          Customer_To_Be_Served);

        Server_Status[1] = Busy;
      }
      else
      {
        Server_Status[1] = Available;
      }
    }
    else
    {
      /*
       * Wait_Queue array starts from 0 based on the structure
       * compared to Server_Status array which starts from 1
       */

      Wait_Queue_Array = Ptr_Customer->Server - 1;
      
      if (Wait_Queue[Wait_Queue_Array].Ptr_Head != NULL)
      {
        Customer_To_Be_Served = Remove_From_Wait_Queue (
                                  &Wait_Queue[Wait_Queue_Array]);        
        Customer_To_Be_Served->Server = Ptr_Customer->Server;
        Customer_Event_Time =
          Event_Time + Customer_To_Be_Served->Parameters.Service_Time;
        Customer_To_Be_Served_Status = Departing;

        Insert_In_Next_Event_List (
          &(*Ptr_Next_Event_List),
          Customer_Event_Time,
          Customer_To_Be_Served_Status,
          Customer_To_Be_Served);

        Server_Status[Ptr_Customer->Server] = Busy;
      }
      else
      {
        Server_Status[Ptr_Customer->Server] = Available;
      }
    }
  }

  return;

} /* Handle_Departing */

