/* Project_2.c  --- EE 368  -- Fall 2000*/

/*   $Author: ee368c $ */
/*     $Date: 2000/10/06 18:44:18 $ */
/* $Revision: 1.1 $ */


#include "P2.h"

void Print_Matrix (
  Matrix X,
  int N)
{
  /* Prints a labeled Matrix to standard output */

  int Row = 0;
  int Column = 0;

  if (N == 0)
  {
    printf (" ---- Empty Matrix ---- \n");
    return;
  }
  printf (" ");

  for (Column = 0; Column < N; ++Column)
  {
    printf (" %2c", Column + 'A');
  }
  printf ("\n");

  for (Row = 0; Row < N; ++Row)
  {
    printf ("%c", Row + 'A');

    for (Column = 0; Column < N; ++Column)
    {
      printf (" %2d", X[Row][Column]);
    }
    printf ("\n");
  }

  return;

} /* Print_Matrix */



void Read_Adjacency_Matrix (
  Matrix Adjacency,
  int *Number_Of_Nodes)
{
  /*
   *  Read the Adjacency Matrix information from standard input
   *  with the number of nodes on the first line.
   *  the entrys are 0 for no connection
   *                 a positive integer for a connection
   *            1 <= Value <= 99
   *
   *  This function aborts the program on any error condition
   */

  int Row = 0;
  int Column = 0;
  int Entry = 0;
  int Return_Value = NO_ERRORS_DETECTED;

  fprintf (stderr, "Enter the number of Nodes on the Graph\n");
  fprintf (stderr, "     1 <= Number_Of_Nodes <= %d\n", MAX_NODES);
  fprintf (stderr, "Enter Number_Of_Nodes: ");
  if ((Return_Value = scanf ("%d", &(*Number_Of_Nodes))) <= 0)
  {
    fprintf (stderr, "ERROR -- Invalid Adjacency Matrix Information\n");
    fprintf (stderr, "         Unable to read number of nodes\n");

    if (Return_Value == EOF)
    {
      fprintf (stderr, "    Return Value: EOF\n");
    }
    else
    {
      fprintf (stderr, "    Return Value: %d\n", Return_Value);
    }

    exit (ERROR_DETECTED);
  }

  fprintf (stderr, "\n");

  if (    (*Number_Of_Nodes < 1)
       || (*Number_Of_Nodes > MAX_NODES))
  {
    fprintf (stderr, "ERROR --- The number of nodes must ");
    fprintf (stderr, "be 1 <= N <= %d\n", MAX_NODES);
    fprintf (stderr, "Number read was: %d\n", *Number_Of_Nodes);
    fprintf (stderr, "ABORTING THE PROGRAM\n");

    exit (ERROR_DETECTED);
  }

  /* Read in the Adjacency Matrix */

  fprintf (stderr, "\n");
  fprintf (stderr, "Enter the values of the adjacency Matrix\n");
  fprintf (stderr, "one row at time with %d items per line\n",
    *Number_Of_Nodes);
  fprintf (stderr, "            0       -- No connection\n");
  fprintf (stderr, "    1 <= Value < %d -- Connection\n", MAX_WEIGHT);
  fprintf (stderr, "\n");

  fprintf (stderr, "  ");
  for (Column = 0; Column < *Number_Of_Nodes; ++Column)
  {
    fprintf (stderr, " %2c", Column + 'A');
  }
  fprintf (stderr, "\n");

  for (Row = 0; Row < *Number_Of_Nodes; ++Row)
  {
    fprintf (stderr, "%c  ", Row + 'A');
    for (Column = 0; Column < *Number_Of_Nodes; ++Column)
    {
      if ((Return_Value = scanf ("%d", &Entry)) <= 0)
      {
        fprintf (stderr, "ERROR -- Invalid Adjacency Matrix\n");

        if (Return_Value == EOF)
        {
          fprintf (stderr, "    Return Value: EOF\n");
        }
        else
        {
          fprintf (stderr, "    Return Value: %d\n", Return_Value);
        }

        exit (ERROR_DETECTED);
      }
      else if (    (Entry < 0)
                || (Entry > MAX_WEIGHT))
      {
        fprintf (stderr, "ERROR -- Invalid data for ");
        fprintf (stderr, "Adjacency Matrix -- \n");
        fprintf (stderr, "   Should be 0 for no connection ");
        fprintf (stderr, " or 1 to %d for a connection: ", MAX_WEIGHT);
        fprintf (stderr, " -- Entry was %d\n", Entry);

        exit (ERROR_DETECTED);
      }
      else
      {
        Adjacency[Row][Column] = Entry;
      }
    }
  }

  fprintf (stderr, "\n");

  return;

} /* Read_Adjacency_Matrix */



void Read_To_End_Of_Line (void)
{
  /* Read line until end of line is read */

  int Char = 0;

  while (    (Char != 10)
          && (Char != EOF))
  {
    Char = getchar ();
  }

  if (Char == EOF)
  {
    fprintf (stderr, "Data error: End of file reached prematurely\n");
    fprintf (stderr, "Aborting the program\n");
    exit (ERROR_DETECTED);
  }

  return;

} /* Read_To_End_Of_Line */



void Read_Reply (int *Reply)
{

  *Reply = getchar ();

  if (*Reply == EOF)
  {
    fprintf (stderr, "Data error: End of file reached prematurely\n");
    fprintf (stderr, "Aborting the program\n");
    exit (ERROR_DETECTED);
  }

  return;

} /* Read_To_End_Of_Line */



void Build_Adjacency_Graph (
  Node_Pointer *Ptr_Graph,
  Matrix Adjacency,
  int Number_Of_Nodes)
{
  /* Builds the Adjacency Graph from the Adjacency Matrix */

  int Node_I = 0;
  int Node_J = 0;
  Arc_Pointer Ptr_Arc = NULL;
  Node_Pointer Ptr_Node = NULL;
  Node_Pointer Node_List[MAX_NODES] = {NULL};

  Delete_Adjacency_Graph (&(*Ptr_Graph));

  /* Build a singly linked list of Nodes */

  for (Node_I = Number_Of_Nodes - 1; Node_I >= 0; --Node_I)
  {
    Ptr_Node = (Node_Pointer) malloc (sizeof (NODE));
    assert (Ptr_Node != NULL);
    Ptr_Node->Ptr_Next_Node = *Ptr_Graph;
    Ptr_Node->Node_Info.Name = Node_I;
    Ptr_Node->Node_Info.Visited = False;
    Ptr_Node->Node_Info.Forward_Sort_Order = 0;
    Ptr_Node->Node_Info.Backward_Sort_Order = 0;
    Ptr_Node->Node_Info.Earliest_Finish_Time = 0;
    Ptr_Node->Node_Info.Latest_Start_Time = 0;
    Ptr_Node->Node_Info.Critical_Node = False;
    Ptr_Node->Ptr_From_List = NULL;
    Ptr_Node->Ptr_To_List = NULL;
    *Ptr_Graph = Ptr_Node;
    Node_List[Node_I] = Ptr_Node;
  }

  /* Build the To list */

  for (Node_I = Number_Of_Nodes - 1; Node_I >= 0; --Node_I)
  {
    Ptr_Node = Node_List[Node_I];

    for (Node_J = Number_Of_Nodes - 1; Node_J >= 0; --Node_J)
    {
      if (Adjacency[Node_I][Node_J] > 0)
      {
        Ptr_Arc = (Arc_Pointer) malloc (sizeof (ARC));
        Ptr_Arc->Ptr_Node = Node_List[Node_J];
        Ptr_Arc->Arc_Value = Adjacency[Node_I][Node_J];
        Ptr_Arc->Ptr_Next_Arc = Ptr_Node->Ptr_To_List;
        Ptr_Node->Ptr_To_List = Ptr_Arc;
      }
    }
  }

  /* Build the From list */

  for (Node_I = Number_Of_Nodes - 1; Node_I >= 0; --Node_I)
  {
    Ptr_Node = Node_List[Node_I];

    for (Node_J = Number_Of_Nodes - 1; Node_J >= 0; --Node_J)
    {
      if (Adjacency[Node_J][Node_I] > 0)
      {
        Ptr_Arc = (Arc_Pointer) malloc (sizeof (ARC));
        Ptr_Arc->Ptr_Node = Node_List[Node_J];
        Ptr_Arc->Arc_Value = Adjacency[Node_J][Node_I];
        Ptr_Arc->Ptr_Next_Arc = Ptr_Node->Ptr_From_List;
        Ptr_Node->Ptr_From_List = Ptr_Arc;
      }
    }
  }

  return;

} /* Build_Adjacency_Graph */



void Delete_Adjacency_Graph (Node_Pointer *Ptr_Graph)
{
  Node_Pointer Ptr_Node = NULL;
  Arc_Pointer Ptr_Arc = NULL;

  while (*Ptr_Graph != NULL)
  {
    Ptr_Node = *Ptr_Graph;
    *Ptr_Graph = (*Ptr_Graph)->Ptr_Next_Node;

    while (Ptr_Node->Ptr_To_List != NULL)
    {
      Ptr_Arc = Ptr_Node->Ptr_To_List;
      Ptr_Node->Ptr_To_List =
        Ptr_Node->Ptr_To_List->Ptr_Next_Arc;
      free (Ptr_Arc);
      Ptr_Arc = NULL;
    }

    while (Ptr_Node->Ptr_From_List != NULL)
    {
      Ptr_Arc = Ptr_Node->Ptr_From_List;
      Ptr_Node->Ptr_From_List =
        Ptr_Node->Ptr_From_List->Ptr_Next_Arc;
      free (Ptr_Arc);
      Ptr_Arc = NULL;
    }
    free (Ptr_Node);
    Ptr_Node = NULL;
  }

  return;

} /*Delete_Adjacency_Graph */



void Print_Weighted_Adjacency_Graph (Node_Pointer Ptr_Graph)
{
  int Count = 0;
  Arc_Pointer Ptr_Arc = NULL;
  Node_Pointer Ptr_Node = NULL;

  if (Ptr_Graph == NULL)
  {
    printf (" ---- NULL GRAPH ---- \n");
    return;
  }

  Ptr_Node = Ptr_Graph;

  while (Ptr_Node != NULL)
  {
    printf ("Node %c ", Ptr_Node->Node_Info.Name + 'A');
    printf (" |   To: ");

    Ptr_Arc = Ptr_Node->Ptr_To_List;
    Count = 0;
    while (Ptr_Arc != NULL)
    {
      ++Count;
      if (Ptr_Arc->Ptr_Node != NULL)
      {
        printf ("   %c", Ptr_Arc->Ptr_Node->Node_Info.Name + 'A');
        printf (": %2d", Ptr_Arc->Arc_Value);
        if ((Count % 5) == 0)
        {
          printf ("\n");
        }
      }
      Ptr_Arc = Ptr_Arc->Ptr_Next_Arc;
      if (    (Count % 5 == 0)
           && (Count > 0)
           && (Ptr_Arc != NULL))
      {
        printf ("        |       ");
      }
      else
      {
        if (    ((Count % 5) != 0)
             && (Ptr_Arc == NULL))
        {
          printf ("\n");
        }
      }
    }

    if (Count == 0)
    {
      printf ("\n");
    }
    printf ("        | From: ");

    Ptr_Arc = Ptr_Node->Ptr_From_List;
    Count = 0;
    while (Ptr_Arc != NULL)
    {
      ++Count;
      if (Ptr_Arc->Ptr_Node != NULL)
      {
        printf ("   %c", Ptr_Arc->Ptr_Node->Node_Info.Name + 'A');
        printf (": %2d", Ptr_Arc->Arc_Value);
        if ((Count % 5) == 0)
        {
          printf ("\n");
        }
      }
      Ptr_Arc = Ptr_Arc->Ptr_Next_Arc;
      if (    (Count % 5 == 0)
           && (Count > 0)
        && (Ptr_Arc != NULL))
      {
        printf ("        |       ");
      }
      else
      {
        if (    ((Count % 5) != 0)
             && (Ptr_Arc == NULL))
        {
          printf ("\n");
        }
      }
    }
    if (Count == 0)
    {
      printf ("\n");
    }

    fflush (stdout);
    Ptr_Node = Ptr_Node->Ptr_Next_Node;
  }

  return;

} /* Print_Weighted_Adjacency_Graph */



void Print_Pert_Summary (Node_Pointer Ptr_Graph)
{
  Node_Pointer Ptr_Node;

  Ptr_Node = Ptr_Graph;

  while (Ptr_Node != NULL)
  {
    printf ("Node %c ", Ptr_Node->Node_Info.Name + 'A');
    printf ("  Earliest Finish Time: %4d  ",
      Ptr_Node->Node_Info.Earliest_Finish_Time);
    printf ("Latest Start Time: %4d  ",
      Ptr_Node->Node_Info.Latest_Start_Time);
    if (Ptr_Node->Node_Info.Critical_Node == True)
    {
      printf ("Critical Node\n");
    }
    else
    {
      printf ("\n");
    }

    fflush (stdout);
    Ptr_Node = Ptr_Node->Ptr_Next_Node;
  }

  return;

} /* Print_Pert_Summary */



int main (void)
{
  int Number_Of_Nodes = 0;
  Boolean Error = False;
  Matrix Pert_Matrix = { { 0 } };
  Node_Pointer Ptr_Graph = NULL;
  Node_Pointer Ptr_Node_P = NULL;
  int Return_Value = NO_ERRORS_DETECTED;

  Set_Limits (CALL_LIMIT, MEMORY_LIMIT);

  Read_Adjacency_Matrix (Pert_Matrix, &Number_Of_Nodes);

  printf ("The Pert_Matrix matrix for the ");
  printf ("%2d node graph\n", Number_Of_Nodes);
  printf ("\n");
  fflush (stdout);

  Print_Matrix (Pert_Matrix, Number_Of_Nodes);

  printf ("\n");
  fflush (stdout);

  Build_Adjacency_Graph (&Ptr_Graph, Pert_Matrix, Number_Of_Nodes);

  Error = False;
  Ptr_Node_P = Ptr_Graph;

  while (Ptr_Node_P != NULL)
  {
    if (    (Ptr_Node_P->Ptr_To_List == NULL)
         && (Ptr_Node_P->Ptr_From_List == NULL))
    {
      Error = True;
      printf ("ERROR: Node %c is isolated\n",
        Ptr_Node_P->Node_Info.Name + 'A');
    }
    Ptr_Node_P = Ptr_Node_P->Ptr_Next_Node;
  }

  printf ("Weighted Adjacency Graph\n\n");
  Print_Weighted_Adjacency_Graph (Ptr_Graph);
  printf ("\n");

  if (Error == True)
  {
    Delete_Adjacency_Graph (&Ptr_Graph);
    printf ("FATAL ERROR: Invalid Pert Graph found\n");
    printf ("Terminating the program\n");
    fflush (stderr);
    exit (ERROR_DETECTED);
  }

  Return_Value = Pert (Ptr_Graph);

  if (Return_Value == NO_ERRORS_DETECTED)
  {
    Delete_Adjacency_Graph (&Ptr_Graph);
  }

  Check_Dynamic_Memory ();

  return Return_Value;

} /* main */

