/* Help Program for homework 1, EE 368c, Spring 2000 */

/*     $Author: ee368c $
 *       $Date: 2000/01/07 20:35:35 $
 *   $Revision: 1.1 $
 */

/* Global information */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#define ERROR_DETECTED ( 1 )
#define NO_ERRORS_DETECTED ( 0 )

struct Tree_Node_Info
{
  int Data;
  struct Tree_Node_Info *Ptr_Up;
  struct Tree_Node_Info *Ptr_Left;
  struct Tree_Node_Info *Ptr_Right;
};

typedef struct Tree_Node_Info TREE_NODE;
typedef TREE_NODE *Tree_Node_Pointer;

struct List_Node_Info
{
  Tree_Node_Pointer Ptr_Tree_Node;
  struct List_Node_Info *Ptr_Left;
  struct List_Node_Info *Ptr_Right;
};

typedef struct List_Node_Info LIST_NODE;
typedef LIST_NODE *List_Node_Pointer;

struct Heap_Info
{
  Tree_Node_Pointer Ptr_Top_Of_Heap;
  List_Node_Pointer Ptr_Tail_Of_Next_Connection_Deque;
};

typedef struct Heap_Info HEAP;
typedef HEAP Priority_Heap;



#include "hw1.c"


void Adjust_Heap_Down (Tree_Node_Pointer Ptr_P)
{
  /*
   *  Moves the data into its proper sorted place
   *  in decreasing order on a path from the top
   *  to a terminal node moving along the path with
   *  the largest values. In case of ties the left
   *  tree branch is followed.
   */

  int Temp_Data = 0;

  while (    (    (Ptr_P->Ptr_Left != NULL)
               && (Ptr_P->Ptr_Left->Data > Ptr_P->Data))
          || (    (Ptr_P->Ptr_Right != NULL)
               && (Ptr_P->Ptr_Right->Data > Ptr_P->Data)))
  {
    if (    (Ptr_P->Ptr_Left != NULL)
         && (Ptr_P->Ptr_Right != NULL))
    {
      if (Ptr_P->Ptr_Left->Data > Ptr_P->Ptr_Right->Data)
      {
        Temp_Data = Ptr_P->Data;
        Ptr_P->Data = Ptr_P->Ptr_Left->Data;
        Ptr_P->Ptr_Left->Data = Temp_Data;
        Ptr_P = Ptr_P->Ptr_Left;
      }
      else
      {
        Temp_Data = Ptr_P->Data;
        Ptr_P->Data = Ptr_P->Ptr_Right->Data;
        Ptr_P->Ptr_Right->Data = Temp_Data;
        Ptr_P = Ptr_P->Ptr_Right;
      }
    }
    else
    {
      if (Ptr_P->Ptr_Left != NULL)
      {
        Temp_Data = Ptr_P->Data;
        Ptr_P->Data = Ptr_P->Ptr_Left->Data;
        Ptr_P->Ptr_Left->Data = Temp_Data;
        Ptr_P = Ptr_P->Ptr_Left;
      }
      else
      {
        Temp_Data = Ptr_P->Data;
        Ptr_P->Data = Ptr_P->Ptr_Right->Data;
        Ptr_P->Ptr_Right->Data = Temp_Data;
        Ptr_P = Ptr_P->Ptr_Right;
      }
    }
  }

  return;

} /* Adjust_Heap_Down */



int Remove_From_Head_Of_List (
  List_Node_Pointer *Ptr_Tail,
  Tree_Node_Pointer *Ptr_Tree_Node)
{
  /* 
   *  Remove the pointer stored in the first (left most) node 
   *  of a doubly linked circular list whose last node (right most) is
   *  pointed to by "*Ptr_Tail"
   *  (*Ptr_Tail)->Ptr_Right points to the first (head) node of the list
   *
   *  This function returns ERROR_DETECTED if any errors are detected
   *  and returns NO_ERRORS_DETECTED if no errors are detected
   */

  List_Node_Pointer Ptr_List_Node_To_Be_Removed = NULL;

  if (*Ptr_Tail == NULL)
  {
    return ERROR_DETECTED;    /* Empty list */
  }
  else
  {
    *Ptr_Tree_Node = (*Ptr_Tail)->Ptr_Right->Ptr_Tree_Node;

    Ptr_List_Node_To_Be_Removed = (*Ptr_Tail)->Ptr_Right;

    if ((*Ptr_Tail)->Ptr_Right != *Ptr_Tail)
    {
      (*Ptr_Tail)->Ptr_Right = Ptr_List_Node_To_Be_Removed->Ptr_Right;
      (*Ptr_Tail)->Ptr_Right->Ptr_Left = *Ptr_Tail;
    }
    else
    {
      *Ptr_Tail = NULL;
    }

    free (Ptr_List_Node_To_Be_Removed);
    Ptr_List_Node_To_Be_Removed = NULL;
  }

  return NO_ERRORS_DETECTED;

} /* Remove_From_Head_Of_List */



int Remove_From_Tail_Of_List (
  List_Node_Pointer *Ptr_Tail,
  Tree_Node_Pointer *Ptr_Tree_Node)
{
  /* 
   *  Removes the pointer stored in the last (right most) node 
   *  of a doubly linked circular list whose last node (right most) is
   *  pointed to by "*Ptr_Tail"
   *  (*Ptr_Tail)->Ptr_Right points to the first (head) node of the list
   *
   *  This function returns ERROR_DETECTED if any errors are detected
   *  and returns NO_ERRORS_DETECTED if no errors are detected
   */

  List_Node_Pointer Ptr_List_Node_To_Be_Removed = NULL;

  if (*Ptr_Tail == NULL)
  {
    return ERROR_DETECTED;   /* Empty list */
  }
  else
  {
    *Ptr_Tree_Node = (*Ptr_Tail)->Ptr_Tree_Node;
    Ptr_List_Node_To_Be_Removed = *Ptr_Tail;

    if ((*Ptr_Tail)->Ptr_Right != *Ptr_Tail)
    {
      *Ptr_Tail = (*Ptr_Tail)->Ptr_Left;

      (*Ptr_Tail)->Ptr_Right = Ptr_List_Node_To_Be_Removed->Ptr_Right;
      (*Ptr_Tail)->Ptr_Right->Ptr_Left = *Ptr_Tail;
    }
    else
    {
      *Ptr_Tail = NULL;
    }

    free (Ptr_List_Node_To_Be_Removed);
    Ptr_List_Node_To_Be_Removed = NULL;
  }

  return NO_ERRORS_DETECTED;

} /* Remove_From_Ptr_Tail_Of_List */



int Insert_At_Head_Of_List (
  List_Node_Pointer *Ptr_Tail,
  Tree_Node_Pointer Ptr_Tree_Node)
{
  /* 
   *  Inserts the pointer before the first (left most) node 
   *  of a doubly linked circular list whose last node (right most) is
   *  pointed to by "*Ptr_Tail"
   *  (*Ptr_Tail)->Ptr_Right points to the first (head) node of the list
   *
   *  This function returns ERRORS_DETECTED if any errors are detected 
   *  and returns NO_ERRORS_DETECTED if no errors are detected
   */

  List_Node_Pointer Ptr_New_List_Node = NULL;

  Ptr_New_List_Node = (List_Node_Pointer) malloc (sizeof (LIST_NODE));

  if (Ptr_New_List_Node == NULL)
  {
    return ERROR_DETECTED;   /* Fatal error -- out of memory */
  }

  Ptr_New_List_Node->Ptr_Tree_Node = Ptr_Tree_Node;

  if (*Ptr_Tail == NULL)
  {
    /* List was empty */

    *Ptr_Tail = Ptr_New_List_Node;
    Ptr_New_List_Node->Ptr_Left = Ptr_New_List_Node;
    Ptr_New_List_Node->Ptr_Right = Ptr_New_List_Node;
  }
  else
  {
    /* Multi-node list */

    Ptr_New_List_Node->Ptr_Left = *Ptr_Tail;
    Ptr_New_List_Node->Ptr_Right = (*Ptr_Tail)->Ptr_Right;
    (*Ptr_Tail)->Ptr_Right->Ptr_Left = Ptr_New_List_Node;
    (*Ptr_Tail)->Ptr_Right = Ptr_New_List_Node;
  }

  return NO_ERRORS_DETECTED;

} /* Insert_At_Head_Of_List */



int Insert_At_Tail_Of_List (
  List_Node_Pointer *Ptr_Tail, 
  Tree_Node_Pointer Ptr_Tree_Node)
{
  /* 
   *  Inserts the pointer after the last (right most) node 
   *  of a doubly linked circular list whose last node (right most) is
   *  pointed to by "*Ptr_Tail"
   *  (*Ptr_Tail)->Ptr_Right points to the first (head) node of the list
   *
   *  This function returns ERRORS_DETECTED if any errors are detected 
   *  and returns NO_ERRORS_DETECTED if no errors are detected
   */

  List_Node_Pointer Ptr_New_List_Node = NULL;

  Ptr_New_List_Node = (List_Node_Pointer) malloc (sizeof (LIST_NODE));

  if (Ptr_New_List_Node == NULL)
  {
    return ERROR_DETECTED;   /* Fatal error -- out of memory */
  }

  Ptr_New_List_Node->Ptr_Tree_Node = Ptr_Tree_Node;

  if (*Ptr_Tail == NULL)
  {
    /* Empty list */

    *Ptr_Tail = Ptr_New_List_Node;
    Ptr_New_List_Node->Ptr_Left = Ptr_New_List_Node;
    Ptr_New_List_Node->Ptr_Right = Ptr_New_List_Node;
  }
  else
  {
    /* Multi-Node List */

    Ptr_New_List_Node->Ptr_Left = *Ptr_Tail;
    Ptr_New_List_Node->Ptr_Right = (*Ptr_Tail)->Ptr_Right;
    (*Ptr_Tail)->Ptr_Right->Ptr_Left = Ptr_New_List_Node;
    (*Ptr_Tail)->Ptr_Right = Ptr_New_List_Node;
    (*Ptr_Tail) = Ptr_New_List_Node;
  }

  return NO_ERRORS_DETECTED;

} /* Insert_At_Tail_Of_List */



int Insert_In_Heap (
  Priority_Heap *Heap,
  int Value)
{
  /*
   *  Inserts the integer "Value" into its proper place
   *  in the priority heap "Heap".
   *  The heap is dynamic with the largest value at the
   *  top of the heap
   *
   *  The heap is maintained as a balanced tree. 
   *
   *  A doubly linked circular list is maintained as a deque
   *  holding pointers to the nodes in the tree which have
   *  NULL left or right pointers.  This next connection list
   *  has a pointer to the tail (right most end) of the list.
   *  New nodes are inserted at the tail end as they are created.
   *  Nodes are extracted from the (left most end) as needed
   *  with first the left and then the right pointers being used
   *  to attach the new node to the tree.
   *
   *  This function returns ERRORS_DETECTED if any errors are detected
   *  and returns NO_ERRORS_DETECTED if no errors are detected
   */

  void Adjust_Heap_Up (Tree_Node_Pointer Ptr_Tree_Root);

  int Remove_From_Head_Of_List(
    List_Node_Pointer *Ptr_Tail,
    Tree_Node_Pointer *Ptr_Tree_Node);

  int Insert_At_Tail_Of_List(
    List_Node_Pointer *Ptr_Tail, 
    Tree_Node_Pointer Ptr_Tree_Node);
 
  Tree_Node_Pointer Ptr_New_Tree_Node = NULL;
  Tree_Node_Pointer Ptr_Parent = NULL;
  int Error_Code = NO_ERRORS_DETECTED;

  /* Create a new node for the heap */

  Ptr_New_Tree_Node = (Tree_Node_Pointer) malloc (sizeof (TREE_NODE));

  if (Ptr_New_Tree_Node == NULL)
  {
    return ERROR_DETECTED;   /* Fatal error -- out of memory */
  }

  Ptr_New_Tree_Node->Data = Value;
  Ptr_New_Tree_Node->Ptr_Left = NULL;
  Ptr_New_Tree_Node->Ptr_Right = NULL;

  /* Insert it into the heap */

  if ((*Heap).Ptr_Top_Of_Heap == NULL)
  {
    /* Empty Heap */

    Ptr_New_Tree_Node->Ptr_Up = NULL;
    (*Heap).Ptr_Top_Of_Heap = Ptr_New_Tree_Node;

    /*
     *  Insert a pointer to the new tree node at the tail
     *  of the next connection node list
     */

    Error_Code = Insert_At_Tail_Of_List (
                   &(*Heap).Ptr_Tail_Of_Next_Connection_Deque,
                   Ptr_New_Tree_Node);

    if (Error_Code != NO_ERRORS_DETECTED)
    {
      return ERROR_DETECTED;
    }
  }
  else
  {
    /*  Extract a pointer to the tree node in the heap
     *  to which this node is to be attached.
     *
     *  This node is located at the head of the 
     *  next connection deque
     */

    Error_Code = Remove_From_Head_Of_List (
                   &(*Heap).Ptr_Tail_Of_Next_Connection_Deque,
                   &Ptr_Parent);

    if (Error_Code != NO_ERRORS_DETECTED)
    {
      return ERROR_DETECTED;         /* Fatal programming error */
    }

    /* Connect new node to parent in the heap */

    Ptr_New_Tree_Node->Ptr_Up = Ptr_Parent;

    if (Ptr_Parent->Ptr_Left == NULL)
    {
      Ptr_Parent->Ptr_Left = Ptr_New_Tree_Node;

      /*
       *  Return pointer to parent node to the
       *  head of the next connection deque so
       *  that the right branch can be used for
       *  the next node inserted in the heap
       */

      Error_Code = Insert_At_Head_Of_List (
                     &(*Heap).Ptr_Tail_Of_Next_Connection_Deque,
                     Ptr_Parent);

      if (Error_Code != NO_ERRORS_DETECTED)
      {
        return ERROR_DETECTED;       /* Fatal error */
      }

      /*
       *  Add a pointer to the new node to the
       *  tail of the next connection node list
       */

      Error_Code = Insert_At_Tail_Of_List (
                     &(*Heap).Ptr_Tail_Of_Next_Connection_Deque,
                     Ptr_Parent->Ptr_Left);

      if (Error_Code != NO_ERRORS_DETECTED)
      {
        return ERROR_DETECTED;       /* Fatal Error */
      }
    }
    else
    {
      Ptr_Parent->Ptr_Right = Ptr_New_Tree_Node;

      /*
       *  Add a pointer to the new node to the
       *  tail of the next connection node list
       */

      Error_Code = Insert_At_Tail_Of_List (
                     &(*Heap).Ptr_Tail_Of_Next_Connection_Deque,
                     Ptr_Parent->Ptr_Right);

      if (Error_Code != NO_ERRORS_DETECTED)
      {
        return ERROR_DETECTED;       /* Fatal Error */
      }
    }

    /* Adjust the heap as needed */

    Adjust_Heap_Up (Ptr_New_Tree_Node);

  }

  return NO_ERRORS_DETECTED;

} /* Insert_In_Heap */



void Adjust_Heap_Up (Tree_Node_Pointer Ptr_Tree_Node_P)
{
  /*
   *  Moves the data into its proper sorted place
   *  in increasing order on the path from its
   *  current location to the Top_Of_Heap 
   */

  int Temp_Data = 0;

  while (    (Ptr_Tree_Node_P->Ptr_Up != NULL)
          && (Ptr_Tree_Node_P->Data > Ptr_Tree_Node_P->Ptr_Up->Data))
  {
    /* Swap the Data */

    Temp_Data = Ptr_Tree_Node_P->Data;
    Ptr_Tree_Node_P->Data = Ptr_Tree_Node_P->Ptr_Up->Data;
    Ptr_Tree_Node_P->Ptr_Up->Data = Temp_Data;

    Ptr_Tree_Node_P = Ptr_Tree_Node_P->Ptr_Up;
  }

  return;

} /* Adjust_Heap_Ptr_Up */




int Is_A_Valid_List (List_Node_Pointer Ptr_List)
{
  /*
   *  Checks for a valid doubly linked circular list with
   *  non NULL pointers stored at each node
   *
   *  This function returns ERROR_DETECTED if any errors are detected
   *  and returns NO_ERRORS_DETECTED if no errors are detected
   */

  List_Node_Pointer Ptr_List_Node_P = Ptr_List;

  if (Ptr_List_Node_P == NULL)
  {
    return NO_ERRORS_DETECTED;   /* Empty list */
  }

  do
  {
    if (    (Ptr_List_Node_P->Ptr_Tree_Node == NULL)
         || (Ptr_List_Node_P->Ptr_Right == NULL)
         || (Ptr_List_Node_P->Ptr_Right->Ptr_Left != Ptr_List_Node_P)
         || (Ptr_List_Node_P->Ptr_Left == NULL)
         || (Ptr_List_Node_P->Ptr_Left->Ptr_Right != Ptr_List_Node_P))
    {
      return ERROR_DETECTED;         /* Invalid list */
    }
    else
    {
      Ptr_List_Node_P = Ptr_List_Node_P->Ptr_Right;
    }
  } while (Ptr_List_Node_P != Ptr_List);

  return NO_ERRORS_DETECTED;

} /* Is_A_Valid_List */



int Print_List_From_Head_To_Tail (List_Node_Pointer Ptr_Tail)
{
  /*
   *  This function prints out the data stored in the node
   *  pointed to by the pointers stored in the nodes of the
   *  doubly linked list whose last (right most) node is pointed
   *  to by "Ptr_Tail".  The first node, head, is pointed to by
   *  "Ptr_Tail->Ptr_Right"
   *
   *  This function returns ERROR_DETECTED if any errors are detected
   *  and returns NO_ERRORS_DETECTED if no errors are detected
   */

  int Is_A_Valid_List (List_Node_Pointer Ptr_List);
  int Error_Code = NO_ERRORS_DETECTED;
  List_Node_Pointer Ptr_List_Node_P = NULL;

  Error_Code = Is_A_Valid_List (Ptr_Tail);

  if (Error_Code != NO_ERRORS_DETECTED)
  {
    printf (" Invalid list detected");
    fflush (stdout);
    return ERROR_DETECTED;
  }
  else
  {
    if (Ptr_Tail == NULL)
    {
      printf (" List is empty");
    }
    else
    {
      Ptr_List_Node_P = Ptr_Tail->Ptr_Right;

      do
      {
        printf (" %3d", Ptr_List_Node_P->Ptr_Tree_Node->Data);

        Ptr_List_Node_P = Ptr_List_Node_P->Ptr_Right;
      } while (Ptr_List_Node_P != Ptr_Tail->Ptr_Right);
    }
  }

  printf ("\n");

  return NO_ERRORS_DETECTED;


} /* Print_List_From_Head_To_Tail */



int Print_Tree (Tree_Node_Pointer Ptr_Tree_Node_P)
{
  int Is_A_Valid_Tree_Structure (
    List_Node_Pointer *Ptr_Tail, Tree_Node_Pointer Ptr_Tree_Node_P);

  void Delete_List (List_Node_Pointer *Ptr_Tail);

  int Is_A_Valid_Heap_Tree (Tree_Node_Pointer Ptr_Tree_Node_P);

  void Print_Tree_LNR (
    Tree_Node_Pointer Ptr_Tree_Node_P,
    int Level);

  List_Node_Pointer Ptr_Tail = NULL;
  int Error_Code = NO_ERRORS_DETECTED;

  Error_Code = Is_A_Valid_Tree_Structure (&Ptr_Tail, Ptr_Tree_Node_P);

  if (Error_Code != NO_ERRORS_DETECTED)
  {
    return Error_Code;
  }

  Delete_List (&Ptr_Tail);

  Error_Code = Is_A_Valid_Heap_Tree (Ptr_Tree_Node_P);

  if (Error_Code != NO_ERRORS_DETECTED)
  {
    return Error_Code;
  }

  if (Ptr_Tree_Node_P == NULL)
  {
    printf ("\nHeap is Empty\n");
  }
  else
  {
    Print_Tree_LNR (Ptr_Tree_Node_P, 0);
  }

  return NO_ERRORS_DETECTED;


} /* Print_Tree */



void Delete_List (List_Node_Pointer *Ptr_Tail)
{
  List_Node_Pointer Ptr_List_Node_P = NULL;
  List_Node_Pointer Ptr_List_Node_to_Be_Deleted = NULL;

  if (*Ptr_Tail == NULL)
  {
    return;
  }
  else
  {
    Ptr_List_Node_P = *Ptr_Tail;

    Ptr_List_Node_P->Ptr_Right->Ptr_Left = NULL;
    Ptr_List_Node_P->Ptr_Right = NULL;

    while (Ptr_List_Node_P != NULL)
    {
      Ptr_List_Node_to_Be_Deleted = Ptr_List_Node_P;
      Ptr_List_Node_P = Ptr_List_Node_P->Ptr_Left;
      free (Ptr_List_Node_to_Be_Deleted);
      Ptr_List_Node_to_Be_Deleted = NULL;
    }
  }

  return;

} /* Delete_List */



void Print_Tree_LNR (
  Tree_Node_Pointer Ptr_Tree_Node_P,
  int Level)
{
  int I = 0;

  if (Ptr_Tree_Node_P != NULL)
  {
    Print_Tree_LNR (Ptr_Tree_Node_P->Ptr_Right, Level + 1);

    for (I = 0; I <= 3*Level; I++)
      printf (" ");
    printf ("%2d\n", Ptr_Tree_Node_P->Data);
    fflush (stdout);

    Print_Tree_LNR (Ptr_Tree_Node_P->Ptr_Left, Level + 1);
  }

  return;

} /* Print_Tree_LNR */



int Is_A_Valid_Tree_Structure (
  List_Node_Pointer *Ptr_Tail,
  Tree_Node_Pointer Ptr_Tree_Node_P)
{
  int Insert_At_Tail_Of_List (
    List_Node_Pointer *Ptr_Tail,
    Tree_Node_Pointer Ptr_Tree_Node);

  int Is_Not_Found_In_List (
    List_Node_Pointer Ptr_Tail,
    Tree_Node_Pointer Ptr_Tree_Node);

  int Error_Code = NO_ERRORS_DETECTED;

  if (Ptr_Tree_Node_P == NULL)
  {
    return NO_ERRORS_DETECTED;
  }

  Error_Code = Is_A_Valid_Tree_Structure (
                 &(*Ptr_Tail), Ptr_Tree_Node_P->Ptr_Left);

  if (Error_Code != NO_ERRORS_DETECTED)
  {
    return Error_Code;
  }
  else
  {
    Error_Code = Is_A_Valid_Tree_Structure (
                   &(*Ptr_Tail), Ptr_Tree_Node_P->Ptr_Right);

    if (Error_Code != NO_ERRORS_DETECTED)
    {
      return Error_Code;
    }
    else
    {
      Error_Code = Is_Not_Found_In_List (*Ptr_Tail, Ptr_Tree_Node_P);

      if (Error_Code != NO_ERRORS_DETECTED)
      {
        return Error_Code;
      }
      else
      {
        Error_Code = Insert_At_Tail_Of_List (
                       &(*Ptr_Tail), Ptr_Tree_Node_P);

        if (Error_Code != NO_ERRORS_DETECTED)
        {
          return Error_Code;
        }
      }
    }
  }

  return NO_ERRORS_DETECTED;

} /* Is_A_Valid_Tree_Structure */


int Is_Not_Found_In_List (
  List_Node_Pointer Ptr_Tail,
  Tree_Node_Pointer Ptr_Tree_Node_P)
{
  List_Node_Pointer Ptr_List_Node_Q = Ptr_Tail;

  if (Ptr_List_Node_Q == NULL)
  {
    return NO_ERRORS_DETECTED;   /* Empty list */
  }

  do
  {
    if (Ptr_List_Node_Q->Ptr_Tree_Node == Ptr_Tree_Node_P)
    {
      return ERROR_DETECTED;         /* Found in the list */
    }
    else
    {
      Ptr_List_Node_Q = Ptr_List_Node_Q->Ptr_Right;
    }

  } while (Ptr_List_Node_Q != Ptr_Tail);

  return NO_ERRORS_DETECTED;     /* Item is not in the list */

} /* Is_Not_Found_In_List */


int Is_A_Valid_Heap_Tree (Tree_Node_Pointer Ptr_Tree_Node_P)
{
  int Check_Heap_Tree (Tree_Node_Pointer Ptr_Tree_Node_P);

  int Error_Code = NO_ERRORS_DETECTED;

  if (Ptr_Tree_Node_P == NULL)
  {
    return NO_ERRORS_DETECTED;   /* Empty heap */
  }
  else
  {
    if (Ptr_Tree_Node_P->Ptr_Up != NULL)
    {
      return ERROR_DETECTED;         /* Invalid heap_tree */
    }
    else
    {
      Error_Code = Check_Heap_Tree (Ptr_Tree_Node_P);

      return Error_Code;
    }
  }

  return NO_ERRORS_DETECTED;

} /* Is_A_Valid_Heap_Tree */



int Check_Heap_Tree (Tree_Node_Pointer Ptr_Tree_Node_P)
{
  int Error_Code = NO_ERRORS_DETECTED;

  if (Ptr_Tree_Node_P == NULL)
  {
    return NO_ERRORS_DETECTED;
  }

  if (Ptr_Tree_Node_P->Ptr_Left != NULL)
  {
    if (Ptr_Tree_Node_P->Ptr_Left->Ptr_Up != Ptr_Tree_Node_P)
    {
      return ERROR_DETECTED;         /* Invaild tree */
    }

    if (Ptr_Tree_Node_P->Ptr_Left->Data > Ptr_Tree_Node_P->Data)
    {
      return ERROR_DETECTED;         /* Invalid heap */
    }
  }

  if (Ptr_Tree_Node_P->Ptr_Right != NULL)
  {
    if (Ptr_Tree_Node_P->Ptr_Right->Ptr_Up != Ptr_Tree_Node_P)
    {
      return ERROR_DETECTED;         /* Invaild tree */
    }

    if (Ptr_Tree_Node_P->Ptr_Right->Data > Ptr_Tree_Node_P->Data)
    {
      return ERROR_DETECTED;         /* Invalid heap */
    }
  }

  Error_Code = Check_Heap_Tree (Ptr_Tree_Node_P->Ptr_Left);

  if (Error_Code != NO_ERRORS_DETECTED)
  {
    Error_Code = Check_Heap_Tree (Ptr_Tree_Node_P->Ptr_Right);
  }

  return Error_Code;

} /* Check_Heap_Tree */


int main (void)
{
  int Insert_In_Heap (
    Priority_Heap *Heap,
    int Value);

  int Print_List_From_Head_To_Tail (List_Node_Pointer Ptr_Tail);

  int Print_Tree (Tree_Node_Pointer Ptr_Tree_Node_P);

  int Remove_From_Heap (
    Priority_Heap *Heap,
    int *Value);

  int I = 0;
  int J = 0;

  Priority_Heap Heap = {NULL, NULL};
  int Error_Code = NO_ERRORS_DETECTED;

  printf ("Testing Insert_In_Heap\n\n");

  printf ("Initial Free Terminal Node List: ");
  Print_List_From_Head_To_Tail (
    Heap.Ptr_Tail_Of_Next_Connection_Deque);
  printf ("\n");
  printf ("Initial Tree\n");
  Error_Code = Print_Tree (Heap.Ptr_Top_Of_Heap);

  if (Error_Code != NO_ERRORS_DETECTED)
  {
    printf ("\nINVALID HEAP DETECTED\n");
  }
  printf ("\n");

  for (I = 1; I <= 15; I++)
  {
    printf ("Inserting %d in the heap\n", I);
    fflush (stdout);
    Error_Code = Insert_In_Heap (&Heap, I);
    printf ("ERROR CODE returned: %d\n", Error_Code);
    printf (" New Free Terminal Node List: ");
    Print_List_From_Head_To_Tail (
      Heap.Ptr_Tail_Of_Next_Connection_Deque);
    printf ("\nRESULTING TREE\n");
    Error_Code = Print_Tree (Heap.Ptr_Top_Of_Heap);

    if (Error_Code != NO_ERRORS_DETECTED)
    {
      printf ("\nINVALID HEAP DETECTED\n");
    }
    printf ("\n");
    fflush (stdout);
  }

  printf ("\nTesting Remove_From_Heap\n\n");

  printf ("Initial Free Terminal Node List: ");
  Print_List_From_Head_To_Tail (
    Heap.Ptr_Tail_Of_Next_Connection_Deque);
  printf ("\n");
  printf ("Initial Tree\n");
  Error_Code = Print_Tree (Heap.Ptr_Top_Of_Heap);

  if (Error_Code != NO_ERRORS_DETECTED)
  {
    printf ("\nINVALID HEAP DETECTED\n");
  }
  printf ("\n");

  for (I = 1; I <= 16; I++)
  {
    printf ("Removing top item from the heap ");
    fflush (stdout);
    Error_Code = Remove_From_Heap (&Heap, &J);
    if (Error_Code != NO_ERRORS_DETECTED)
    {
      printf ("\n");
    }
    else
    {
      printf ("Item removed was %d\n", J);
    }
    printf ("ERROR CODE returned: %d\n", Error_Code);
    fflush (stdout);
    printf (" New Free Terminal Node List: ");
    Print_List_From_Head_To_Tail (
      Heap.Ptr_Tail_Of_Next_Connection_Deque);
    printf ("\nRESULTING TREE\n");
    Error_Code = Print_Tree (Heap.Ptr_Top_Of_Heap);

    if (Error_Code != NO_ERRORS_DETECTED)
    {
      printf ("\nINVALID HEAP DETECTED\n");
    }
    printf ("\n");
    fflush (stdout);
  }

  return NO_ERRORS_DETECTED;     /* No errors detected */

} /* main */